/*
 *
 * Copyright  2004, University of Washington, 
 * Department of Computer Science and Engineering. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *
 * 3. Neither name of the University of Washington, Department of Computer 
 * Science and Engineering nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR 
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 *
 */

package cse466;

import cse466.parser.*;
import cse466.actions.*;
import cse466.packets.*;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;

/**

This is the main class for the tester.  It controls the parsing and
running of the tests.  At this point, it just logs what happens to
System.out, but it would probably be nice to add some other type of
checking later.  This tester is designed to be used with the
net.tinyos.sf.SerialForwarder to do the serial communication with the
mote.

*/
public class Tester {
  /** The socket to the SerialForwarder */
  protected Socket _sock;

  /** The output stream to the SerialForwarder */
  protected OutputStream _sockOs;

  /** The input stream from the SerialForwarder */
  protected InputStream _sockIs;

  /** 

  A lock to prevent interleaved writes to System.out  There
  is the potential since the receiving from the SerialForwarder and
  sending to the SerialForwarder are on separate threads
  
  */
  protected Object _stdOutLock;
  
  /** The time that this test was started.  

  This is used to record the time after the start that events came in

  */
  protected long _startTime;
  
  /** My singleton instance */
  static protected Tester _self = null;

  /** This class is a singleton, so this is protected */
  protected Tester(){
    _sockOs = null;
    _sockIs = null;
    _stdOutLock = new Object();
    _self = this;
  }

  /** Get the singleton instance of the tester */
  public static Tester getInstance(){
    if(_self == null){
      _self = new Tester();
    }
    return _self;
  }

  /**

  Parses the input file that should contain a list of actions.  

  The actions can be either a delay or a packet.  The delay will just
  delay the sending thread by the specified number of milliseconds.
  The packet action will send the specified packet.  The packet
  specification looks like this:

  <pre><code>
  packet &lt;<i>packetType</i>&gt;
  &lt;<i>fieldName</i>&gt; &lt;<i>value</i>&gt;
  &lt;<i>fieldName</i>&gt; &lt;<i>value</i>&gt;
  end
  </code></pre>

  And it must have as many fieldNames as are in the packet.  The delay
  is simple:

  <pre><code>
  delay &lt;<i>delayInMilliseconds</i>&gt;
  </pre></code>

  @param filename The name of the file that contains the actions.

  @param ignoreBadAMValues if false, it will validate the am values of
  the packets

  @return The list of actions that are defined in the file

  @throws ParserException if the input is malformed.

  @throws IOException if there was a problem reading from the file.

  */
  public List parseFile(String filename,
                        boolean ignoreBadAMValues) 
    throws ParserException, IOException {
    
    return Parser.parseFile(filename, ignoreBadAMValues);
  }

  /**

  Sets up the connection to the SerialForwarder.

  @param host the name of the host where the SerialForwarder is
  running

  @param port the port on which the SerialForwarder is listening

  @throws IOException if there was a problem connecting to the
  SerialForwarder

  */
  public void setupSocket(String host, int port) throws IOException {
    _startTime = System.currentTimeMillis();
    _sock = new Socket(host, port);
    _sockOs = _sock.getOutputStream();
    _sockIs = _sock.getInputStream();
    
  }

  /**

  Run through the list and do the actions

  @param actionsList The list of actions to perform.  The run method will
  be called on each one in order
  
  @throws IOException if there was a problem writing to the SerialForwarder.

  */
  public void run(List actionsList) throws IOException {
    byte VERSION[] = {'T', ' '};
	byte PLATFORM[] = {1, 0, 0, 0};
	
	  // send handshake
    _sockOs.write(VERSION);
    _sockOs.write(PLATFORM);
  
	  // give time for the handshake response
	  try {
		  Thread.sleep(100);
	  } catch (Exception e){
		  System.out.println("Sleep Problem");
	  }

   // get the response of the handshake
	byte receivedBytes[] = new byte[6];
    if(_sockIs.read(receivedBytes, 0, 6) != 6)
		System.out.println("DID NOT GET ALL THE SYNC BYTES!");

	if(receivedBytes[0] != 'T' || receivedBytes[1] != ' ') 
	  {
		  _sock.close();
		  throw new IOException("Got bad handshake");
	  }

	// start the receiver running
	(new Thread(new SocketReader())).start();

    // go through the list and call run on each one in order.
    ListIterator i = actionsList.listIterator();
    while(i.hasNext()){
      ((Action)i.next()).run(_sockOs);
    } 
  }



  public static void printUsage(){
    System.err.println("Usage: Tester <filename> <host> <port> [-i]");
  }



  /**
  
  Log to System.out that a packet was received

  */
  public void logReceivedPacket(CSE466Packet pkt){
    long now = System.currentTimeMillis() - _startTime;
    String name = pkt.getClass().getName();
    synchronized(_stdOutLock){
      System.out.println("RECEIVED a " + name + " at " + now);
      pkt.print();
      System.out.println();
    }
  }

  /**
     
  Log to System.out that a packet was sent

  */
  public void logSentPacket(CSE466Packet pkt){
    long now = System.currentTimeMillis() - _startTime;
    String name = pkt.getClass().getName();
    synchronized(_stdOutLock){
      System.out.println("SENT a " + name + " at " + now);
      pkt.print();
      System.out.println();
    }
  }



  public static void main(String[] args){
    try {
      String filename = args[0];
      String host = args[1];
      int port = Integer.parseInt(args[2].trim());
      boolean ignoreBadAMValues;
      if(args.length == 4 && args[3].equals("-i")){
        ignoreBadAMValues = true;
      } else {
        ignoreBadAMValues = false;
      }
      

      List commands = 
        Tester.getInstance().parseFile(filename, 
                                       ignoreBadAMValues);
      Tester.getInstance().setupSocket(host, port);
      Tester.getInstance().run(commands);
    } catch (NumberFormatException nfe){
      System.err.println("Port must be an integer");
      //nfe.printStackTrace();
      printUsage();
    } catch (ArrayIndexOutOfBoundsException aioobe){
	//aioobe.printStackTrace();
      printUsage();
    } catch (ParserException pe) {
	//pe.printStackTrace();
      System.err.println(pe);
    } catch (IOException ioe) {
	//ioe.printStackTrace();
      System.err.println(ioe);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
   

  /**
     
  Reads from the socket and logs when packets come in.

  */
  protected class SocketReader implements Runnable {
    public SocketReader() {
    }
    
    public void run(){

      // here are a bunch of packets that we'll just keep setting to 
      // the incoming array of bytes
      CSE466Packet receivedPacket = null;
      CSE466Packet cse466packet = 
        new CSE466Packet();
      AdjustGlobalsPacket adjustGlobalsPacket = 
        new AdjustGlobalsPacket();
      StopAndListenPacket stopAndListenPacket = 
        new StopAndListenPacket();
      SangPhrasesPacket sangPhrasesPacket =
	new SangPhrasesPacket();
      AlteredGenesPacket alteredGenesPacket =
	new AlteredGenesPacket();
      StartledPacket startledPacket =
	  new StartledPacket();  
      /*      SingSongNPacket singSongNPacket =
	      new SingSongNPacket();*/
      LedsPacket ledsPacket = 
        new LedsPacket();
      LightSensorPacket lightSensorPacket = new LightSensorPacket();

      try{
		  for(;;)
		  {
			  // the first byte tells us the size
			  int size = _sockIs.read();
			  if(size < 5)
				  System.out.println("Extra Byte, Sync Problem" + Byte.toString((byte)size));
			  else
			  {
				  byte[] bytes = new byte[size + 1];
				  bytes[0] = (byte)size;

				  // read size-1 bytes and stick them into the array
				  for(int i=1; i<size+1; i++)
				  {
					  bytes[i] = (byte)_sockIs.read();
				  }

				  // figure out what type of packet we have, and update that packet
				  cse466packet.setFromBytes(bytes);
				  switch(cse466packet.amType)
				  {
					  case 4:
						  ledsPacket.setFromBytes(bytes);
						  receivedPacket = ledsPacket;
						  break;
					  case 20:
						  lightSensorPacket.setFromBytes(bytes);
						  receivedPacket =lightSensorPacket;
						  break;
					  case 50:
						  adjustGlobalsPacket.setFromBytes(bytes);
						  receivedPacket = adjustGlobalsPacket;
						  break;
					  case 51:
						  stopAndListenPacket.setFromBytes(bytes);
						  receivedPacket = stopAndListenPacket;
						  break;
					  case 42:
						  sangPhrasesPacket.setFromBytes(bytes);
						  receivedPacket = sangPhrasesPacket;
						  break;
				          case 43:
					          alteredGenesPacket.setFromBytes(bytes);
						  receivedPacket = alteredGenesPacket;
						  break;
						  /*					  case 52:
						  singSongNPacket.setFromBytes(bytes);
						  receivedPacket = singSongNPacket;
						  break;*/
				          case 60:
					          startledPacket.setFromBytes(bytes);
					          receivedPacket = startledPacket;
						  break;
					  default:
						  synchronized(_stdOutLock)
						  {
							  System.out.println("ERROR: unknown packet type received: " +
								  cse466packet.amType);
							  for(int i=0; i<bytes.length; i++)
							  {
								  System.out.print(bytes[i] + " ");
							  }
							  System.out.println();
						  }
						  receivedPacket = null;
				  }

				  // if we got a recognized packet, log it
				  if(receivedPacket != null) 
				  {
					  logReceivedPacket(receivedPacket);
				  }
			  }
		  }
      } catch (IOException ioe) {
        // lets hope this never happens...
        ioe.printStackTrace();
        try {
          _sock.close();
        } catch (IOException ioe2) {
          // we're already pretty screwed up so there's nothing to do 
          // to try and recover.
        }
      }
    }
  }
}

