/*
 *
 * Copyright  2005, University of Washington, 
 * Department of Computer Science and Engineering. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *
 * 3. Neither name of the University of Washington, Department of Computer 
 * Science and Engineering nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR 
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 *
 */

package cse466.packets;

/**

The base class for all of the packets.  It contains the fields that
are in the TOS_Msg header.

Note: The parser uses the class's fields to figure out how to parse
the file, so if you add or remove fields, you will confuse the parser.

*/
public class CSE466Packet {
  /** where this packet is headed */
  public int destination;

  /** the amType of this packet */
  public int amType;

  /** the groupID of this packet */
  public int groupId;

  
  public CSE466Packet(){
    groupId = 42; // just some magical groupId
    destination = 0xffff; // broadcast
    amType = 4; // leds
  }


  /**

  Returns the proper AMType for this packet, or -1 if the packet doesn't
  have a proper amType.

  */
  public int getProperAMType(){
    return -1;
  }

  /**

  Sets the fields in this packet from the byte array given.  It assumes that
  this packet's fields are at the very front of the array.  It then will
  return its payload.

  */
  public byte[] setFromBytes(byte[] array){
    destination = bytesToInt(array[1], array[2]);
    amType = byteToInt(array[3]);
    groupId = byteToInt(array[4]);
    return subArray(array,6);
  }

  /**

  Gives you the byte array that corresponds to this packet.  The child
  packet is the payload of this packet.  The returned array contains this
  packet including the payload
     
  @param childPacket the payload.  This will be put on the end of the
  array

  @return an array representing this packet
  */
  protected byte [] getByteArray(byte [] childPacket){
    byte [] ret = new byte[childPacket.length + 6];
    ret[0] = intToByte(childPacket.length + 5);
    ret[1] = intToLSB(destination);
    ret[2] = intToMSB(destination);
    ret[3] = intToByte(amType);
    ret[4] = intToByte(groupId);
    ret[5] = intToByte(childPacket.length);
    copyIntoArray(ret, childPacket, 6);
    return ret;
  }

  /**

  Gives you the byte array corresponding to this packet for sending
  over the network.

  */
  public byte [] getByteArray() {
    return getByteArray(new byte[0]);
  }


  /**

  Pretty print the contents of this packet to System.out

  */
  public void print(){
    System.out.println("cse466 destination         : " + Integer.toHexString(destination));
    System.out.println("cse466 amType              : " + Integer.toHexString(amType));
    System.out.println("cse466 groupId             : " + Integer.toHexString(groupId));
  }


  /**

  Returns true if the fields of this packet match the other packet's fields

  @param other The packet to compare to.

  */
  public boolean equals(CSE466Packet other){
    if(destination != other.destination) return false;
    if(amType      != other.amType     ) return false;
    if(groupId     != other.groupId    ) return false;
    return true;
  }




  /**************************************************************
   *
   * Utility functions below here
   *
   **************************************************************/

  /**
     
  Gives you a sub array starting at startIndex in original and going
  to the end.

  */
  protected byte[] subArray(byte[] original, 
                            int startIndex){
    byte[] ret = new byte[original.length - startIndex];
    for(int i=0; i<original.length - startIndex; i++){
      ret[i] = original[i+startIndex];
    }
    return ret;
  }

  /**

  Copies the contents of fromHere into intoHere starting at an offset
  of offsetIntoIntoHere, and going for as long as the fromHere array is.

  */
  protected void copyIntoArray(byte[] intoHere,
                               byte[] fromHere,
                               int offsetIntoIntoHere){
    for(int i=0; i<fromHere.length; i++){
      intoHere[offsetIntoIntoHere+i] = fromHere[i];
    }
  }





  /**
     Get the least significant byte of the value
   */
  protected byte intToLSB(int value) {
    return (byte)(value & 0xff);
  }

  /**
     Get the least significant byte of the value
   */
  protected byte intToMSB(int value) {
    return (byte)((value >> 8) & 0xff);
  }

  /**
     Get the byte value of the value
  */
  protected byte intToByte(int value) {
    return (byte)value;
  }

  /**
     Get the int value of the two bytes.

     @param first The least significant byte

     @param second The most significant byte
  */
  protected int bytesToInt(byte first, byte second){
    return 
      (((int)first) & 0xff) | 
      ((((int)second) << 8) & 0xff00);
  }

  /**
     Get the int value of the byte
  */
  protected int byteToInt(byte value){
    return (int)value & 0xff;
  }
}
