/*
 *
 * Copyright  2004, University of Washington, 
 * Department of Computer Science and Engineering. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *
 * 3. Neither name of the University of Washington, Department of Computer 
 * Science and Engineering nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR 
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 *
 */

package cse466.parser;

import cse466.packets.*;
import cse466.actions.*;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/**

   This class implements the functionality needed for parsing the
   input file and generating the action list that will later be
   executed.

*/
public class Parser {
  /** The current line number being parsed */
  protected int _lineNumber;
  /** The buffered reader that is reading the file */
  protected BufferedReader _br;

  /**
    
  Initializes the current line, and assigns the buffered reader to be
  the one the rest of the parser will use

  @param br The buffered reader that will be used for all the file
  reading

 */
  protected Parser(BufferedReader br){
    _lineNumber = 0;
    _br = br;
  }

  /**

  A helper function that splits a string into two space delimited
  strings and then trims them.  It will throw a parser exception if it
  can't split the string.

  @param s the string to split

  @return a two element array of strings that are the result of
  trimming and splitting the input string

  @throws ParserException if it cannot figure out how to split the
  string

   */ 
  protected String[] splitInTwo(String s) throws ParserException {
    
    String[] ret = s.trim().split("\\s+");
    if(ret.length != 2){
      throw new ParserException("Expected two strings", _lineNumber);
    }
    ret[0] = ret[0].trim();
    ret[1] = ret[1].trim();
    return ret;
  }

  /**

  Grab the next line from the file.  This one allows the end of the
  file to be reached, and returns null in that case.

  @return the next line in the file

  @throws IOException if one occurs from trying to read the file

  */
  protected String nextLineEndOk() throws IOException {
    String ret;
    do{
      _lineNumber++;
      ret = _br.readLine();
      if(ret == null) break;
      ret = ret.trim();
    } while (ret.equals("") || ret.startsWith("#")); // eat empty lines
    
    return ret;
  }

  /**

  Grab the next line from the file, but if it hits the end of the file
  it will throw an IOException saying that it was expecting more.
  This is useful if you're in the middle of parsing a multi-line
  structure.

  @return The next line in the file

  @throws IOException if the end of the file was reached or if there
  was a problem trying to read the file.

  */
  protected String nextLine() throws IOException {
    String s = nextLineEndOk();
    if(s == null){
      throw new IOException("Unexpected end of file");
    }
    return s;
  }
  
  /**
     
  Creates a DelayAction out of the input line.  Since the delay action
  fits on a single line, it does not do any file reading of its own. The 
  format of the delay action in the file is:
  <pre><code>
  delay <i>timeInMilliseconds</i>
  </code></pre>

  @param line The line of the file that contains the delay
  specification.

  @return The Action that represents the delay

  @throws ParserException if there was something wrong with the line
  of input.

  */
  protected Action parseDelayAction(String line) throws ParserException {
    long delay;
    String[] tokens = splitInTwo(line);
    try {
      delay = Long.parseLong(tokens[1]); 
    } catch (NumberFormatException nfe) {
      throw new ParserException("Delay must be a number", _lineNumber);
    }
    return new DelayAction(delay);
  }

  /**

  Parses the sequence of lines that make up a @link PacketAction.  A
  PacketAction simply consists of a packet to send.  These packets can
  be any of the packets in the cse466.packets package.  The format of
  a packet action is:

  <pre><code>
  packet <i>packetType</i>
  <i>fieldName</i> <i>value</i>
  <i>fieldName</i> <i>value</i>
  end
  </code></pre>

  Note that you must have as many fieldname/value pairs as the packet
  has fields.

  @param line The line that contains the start of the packet action
  (packet <i>packetType</i>).

  @param ignoreBadAMValues if true, it will let you put whatever
  amType you want to put in the packet.  If false, it will throw a
  @link ParserException if you try to give a packet an invalid amType.

  @return The action that represents sending the given packet

  @throws ParserException if the input was malformed

  @throws IOException if there was a problem reading from the file.

  */
  protected Action parsePacketAction(String line, boolean ignoreBadAMValues) 
    throws ParserException, IOException {
    
    String[] tokens = splitInTwo(line);
    String className = tokens[1];
    
    // this will be to make sure that all the fields were initialized
    HashMap hm = new HashMap();

    // the action that will be ultimately returned
    Action ret = null;
    try {
      // get the class for the requested packet and instantiate it
      Class c = Class.forName("cse466.packets." + className);
      CSE466Packet p = (CSE466Packet)c.newInstance();

      // get the fields of the requested packet and put them into the hashmap
      Field[] fields = c.getFields();
      for(int i=0; i<fields.length; i++){
        hm.put(fields[i].getName(), fields[i]);
      }

      // get the first line of the packet
      line = nextLine();

      // keep going until we hit the end of the packet
      while(!line.equals("end")){
        tokens = splitInTwo(line);
        
        // grab the field that is assigned to on this line
        Field f = (Field)hm.remove(tokens[0]);

        // either the class doesn't have the requested field, or
        // it's already been assigned to in this packet
        if(f == null){
          throw new ParserException(tokens[0] + 
                                    " either already set for object, " +
                                    "or is not a member of packet " +
                                    className, _lineNumber);
        }

        // Java doesn't automatically parse hex, so I have to do it
        // manually...
        if(tokens[1].startsWith("0x")){
          tokens[1] = tokens[1].substring(2);
          f.setInt(p, Integer.parseInt(tokens[1], 16));
        } else {
          f.setInt(p, Integer.parseInt(tokens[1]));
        }

        // check that the amType is good
        if(!ignoreBadAMValues){
          if(tokens[0].equals("amType")){

            // a properAMType less or equal to 0 means that there is not
            // a defined amType for this type of packet
            if(p.getProperAMType() > 0 && p.amType != p.getProperAMType()){
              throw new ParserException("Bad amType: " + p.amType + 
                                        " for " + className, 
                                        _lineNumber);
            } 
          }
        }

        // and finally, grab the next line, but eat white space
        line = nextLine();
      }

      // if there's anything left in the hashmap, that means that something
      // didn't get assigned to in the packet definition.  Alert the user.
      if(hm.size() != 0) {
        throw new ParserException("Field " + 
                                  hm.keySet().toArray()[0].toString() +
                                  " not set in " + className,
                                  _lineNumber);
      }

      
      ret = new PacketAction(p);
    } catch (ClassNotFoundException cnfe) {
      throw new ParserException(className + " is not a known packet type",
                                _lineNumber);
    } catch (InstantiationException ie) {
      throw new ParserException(className + " is not a known packet type",
                                _lineNumber);
    } catch (IllegalAccessException iae) {
      // this should really never happen, so just pass the exception up.
      throw new ParserException(iae.toString(), _lineNumber);
    } catch (NumberFormatException nfe) {
      throw new ParserException("Expected numerical argument", 
                                _lineNumber);
    }
    return ret;
  }
  
  /**

  Grab the next line and figure out what type of action it is (delay or packet)
  then delegate the rest of the parsing to the appropriate method.

  @param ignoreBadAMValues if true, will let you use whatever am
  values you want otherwise, it forces you to use the ones that
  correspond to the packets.

  @return The action that should be run for this line

  @throws ParserException if the input file is malformed

  @throws IOException if there was a problem reading from the file

  */
  protected Action parseNextAction(boolean ignoreBadAMValues) 
    throws IOException, ParserException {
    
    String line = nextLineEndOk();
    
    Action a;
    if(line == null){
      a = null;
    } else if(line.startsWith("packet")){
      a = parsePacketAction(line, ignoreBadAMValues);
    } else if(line.startsWith("delay")){
      a = parseDelayAction(line);
    } else {
      throw 
        new ParserException("Parser expected packet or delay," +
                            " but got \"" + line + "\"", _lineNumber);
    }
    return a;
  }


  /**

  Runs the parser and appends the incoming actions to the list of
  actions.

  @param ignoreBadAMValues if it should ignore when the amValue does
  not match the amValue that the packet is supposed to have

  @return the ordered list of actions to be executed

  @throws ParserException on malformed input

  @throws IOException if there was a problem reading from the file

  */
  protected List doParsing(boolean ignoreBadAMValues) 
    throws IOException, ParserException {
    
    ArrayList ret = new ArrayList();
    Action a = parseNextAction(ignoreBadAMValues);
    while(a != null){
      ret.add(a);
      a = parseNextAction(ignoreBadAMValues);
    }
    return ret;
  }


  /**
     
  Parses the file specified and gives back the ordered list of actions
  declared in the file.

  @param filename The name of the file to parse

  @param ignoreBadAMValues if true, it will ignore the case where the
  packet is receiving an amValue that does not correspond to the type
  of packet being created.

  @return The list of actions to be executed in order

  @throws IOException if there was a problem reading the file

  @throws ParserException if there was malformed input

  */
  public static List parseFile(String filename, boolean ignoreBadAMValues) 
    throws IOException, ParserException{
    
    Parser p = new Parser(new BufferedReader(new FileReader(filename)));
    return p.doParsing(ignoreBadAMValues);
  }
}
